@@ -26,6 +26,7 @@
 #include <stdlib.h>
 #include <string.h>
 #include <glib.h>
+#include <glib-unix.h>
 #include <glib/gprintf.h>
 
 #define N_(x) (x)
@@ -148,6 +149,19 @@ handle_closed (NotifyNotification *notify,
         g_main_loop_quit (loop);
 }
 
+static gboolean
+on_sigint (gpointer data)
+{
+        NotifyNotification *notification = data;
+
+        g_printerr ("Wait cancelled, closing notification\n");
+
+        notify_notification_close (notification, NULL);
+        g_main_loop_quit (loop);
+
+        return FALSE;
+}
+
 static void
 handle_action (NotifyNotification *notify,
                char               *action,
@@ -192,9 +206,33 @@ server_has_capability (const gchar *capability)
         return supported;
 }
 
+/* The XDG Desktop Notifications Specification requires valid UTF-8 for certain
+ * strings. Given the stability/security implications by accepting console
+ * input, we will insist upon valid UTF-8 being provided for these strings, and
+ * print an error message otherwise.
+ */
+static void
+validate_utf8_or_die (const char *str, const char *param)
+{
+        if (! g_utf8_validate (str, -1, NULL))
+        {
+                /* Translators: the parameter is the 'Component' as defined by the
+                 * XDG notification specification (eg, 'Summary', 'Body', etc.) -
+                 * these are separately translated.
+                 */
+                g_printerr (N_("Invalid UTF-8 provided for parameter: %s\n"), param);
+                exit (1);
+        }
+}
+
 int
 main (int argc, char *argv[])
 {
+#ifdef g_autoptr
+        g_autoptr (NotifyNotification) notify = NULL;
+#else
+        NotifyNotification *notify = NULL;
+#endif
         static const char  *summary = NULL;
         char               *body;
         static const char  *type = NULL;
@@ -203,6 +241,10 @@ main (int argc, char *argv[])
         static char       **n_text = NULL;
         static char       **hints = NULL;
         static char       **actions = NULL;
+        static char        *server_name = NULL;
+        static char        *server_vendor = NULL;
+        static char        *server_version = NULL;
+        static char        *server_spec_version = NULL;
         static gboolean     print_id = FALSE;
         static gint         notification_id = 0;
         static gboolean     do_version = FALSE;
@@ -211,7 +253,6 @@ main (int argc, char *argv[])
         static gboolean     wait = FALSE;
         static glong        expire_timeout = NOTIFY_EXPIRES_DEFAULT;
         GOptionContext     *opt_ctx;
-        NotifyNotification *notify;
         GError             *error = NULL;
         gboolean            retval;
 
@@ -225,7 +266,7 @@ main (int argc, char *argv[])
                  ("Specifies the timeout in milliseconds at which to expire the "
                   "notification."), N_("TIME")},
                 {"app-name", 'a', 0, G_OPTION_ARG_STRING, &app_name,
-                 N_("Specifies the app name for the icon"), N_("APP_NAME")},
+                 N_("Specifies the app name for the notification"), N_("APP_NAME")},
                 {"icon", 'i', 0, G_OPTION_ARG_FILENAME, &icon_str,
                  N_("Specifies an icon filename or stock icon to display."),
                  N_("ICON")},
@@ -265,10 +306,6 @@ main (int argc, char *argv[])
 
         setlocale (LC_ALL, "");
 
-#ifndef GLIB_VERSION_2_36
-        g_type_init ();
-#endif
-
         g_set_prgname (argv[0]);
         g_log_set_always_fatal (G_LOG_LEVEL_ERROR | G_LOG_LEVEL_CRITICAL);
 
@@ -289,8 +326,14 @@ main (int argc, char *argv[])
                 exit (0);
         }
 
-        if (n_text != NULL && n_text[0] != NULL && *n_text[0] != '\0')
+        if (n_text != NULL && n_text[0] != NULL)
+        {
                 summary = n_text[0];
+                /* Translators: XDG notification component to be translated
+                 * consistently as within the specification.
+                 */
+                validate_utf8_or_die (summary, N_("Summary"));
+        }
 
         if (summary == NULL) {
                 fprintf (stderr, "%s\n", N_("No summary specified."));
@@ -300,6 +343,11 @@ main (int argc, char *argv[])
         if (n_text[1] != NULL) {
                 body = g_strcompress (n_text[1]);
 
+                /* Translators: XDG notification component to be translated
+                 * consistently as within the specification.
+                 */
+                validate_utf8_or_die (body, N_("Body"));
+
                 if (n_text[2] != NULL) {
                         fprintf (stderr, "%s\n",
                                  N_("Invalid number of options."));
@@ -310,6 +358,18 @@ main (int argc, char *argv[])
         if (!notify_init ("notify-send"))
                 exit (1);
 
+        notify_get_server_info (&server_name,
+                                &server_vendor,
+                                &server_version,
+                                &server_spec_version);
+
+        g_debug ("Using server %s %s, v%s - Supporting Notification Spec %s",
+                 server_name, server_vendor, server_version, server_spec_version);
+        g_free (server_name);
+        g_free (server_vendor);
+        g_free (server_version);
+        g_free (server_spec_version);
+
         notify = g_object_new (NOTIFY_TYPE_NOTIFICATION,
                                "summary", summary,
                                "body", body,
@@ -439,16 +499,20 @@ main (int argc, char *argv[])
         if (print_id) {
                 g_object_get (notify, "id", &notification_id, NULL);
                 g_printf ("%d\n", notification_id);
+                fflush (stdout);
         }
 
         if (wait) {
+                g_unix_signal_add (SIGINT, on_sigint, notify);
                 loop = g_main_loop_new (NULL, FALSE);
                 g_main_loop_run (loop);
                 g_main_loop_unref (loop);
                 loop = NULL;
         }
 
-        g_object_unref (G_OBJECT (notify));
+#ifndef g_autoptr
+        g_clear_object (&notify);
+#endif
 
         notify_uninit ();
 
